use crate::{
    handlers::{utils::package_handler, MainDialogue, State},
    models::{DecryptedAccount, User},
};
use futures::TryStreamExt;
use itertools::Itertools;
use sea_orm::prelude::*;
use serde_json::from_slice;
use std::sync::Arc;
use teloxide::{adaptors::Throttle, net::Download, prelude::*};
use tokio::task::{spawn_blocking, JoinSet};

async fn get_master_pass(
    bot: Throttle<Bot>,
    msg: Message,
    db: DatabaseConnection,
    dialogue: MainDialogue,
    master_pass: String,
    previous: Message,
    accounts: Vec<DecryptedAccount>,
) -> crate::Result<()> {
    let _ = bot.delete_message(previous.chat.id, previous.id).await;
    let user_id = msg.from().unwrap().id.0;
    let master_pass: Arc<str> = master_pass.into();
    let mut join_set = JoinSet::new();
    let mut failed = Vec::new();
    for account in accounts {
        let master_pass = Arc::clone(&master_pass);
        let db = db.clone();
        join_set.spawn(async move {
            let name = account.name.clone();
            let account =
                match spawn_blocking(move || account.into_account(user_id, &master_pass)).await {
                    Ok(Ok(account)) => account,
                    _ => return Err(name),
                };
            account.insert(&db).await.map_err(|_| name)?;
            Ok(())
        });
    }
    while let Some(result) = join_set.join_next().await.transpose()? {
        match result {
            Ok(()) => (),
            Err(name) => failed.push(name),
        }
    }
    if failed.is_empty() {
        bot.send_message(msg.chat.id, "Success").await?;
    } else {
        let text = format!(
            "Failed to create the following accounts:\n{}",
            failed.into_iter().format("\n")
        );
        bot.send_message(msg.chat.id, text).await?;
    }
    dialogue.exit().await?;
    Ok(())
}

async fn get_document(
    bot: Throttle<Bot>,
    msg: Message,
    _: DatabaseConnection,
    dialogue: MainDialogue,
    previous: Message,
) -> crate::Result<()> {
    let _ = bot.delete_message(previous.chat.id, previous.id).await;
    let document = match msg.document() {
        Some(doc) => doc,
        None => {
            bot.send_message(msg.chat.id, "You didn't send a file")
                .await?;
            dialogue.exit().await?;
            return Ok(());
        }
    };
    match document.file_name {
        Some(ref name) if name.trim().ends_with(".json") => (),
        _ => {
            bot.send_message(msg.chat.id, "Invalid file name").await?;
            dialogue.exit().await?;
            return Ok(());
        }
    }
    let mut data = Vec::with_capacity(document.file.size as usize);
    let file = bot.get_file(&document.file.id).await?;
    bot.download_file_stream(&file.path)
        .try_for_each(|bytes| {
            data.extend(bytes);
            async { Ok(()) }
        })
        .await?;
    let accounts = match spawn_blocking(move || from_slice::<User>(&data)).await? {
        Ok(user) => user.accounts,
        Err(_) => {
            bot.send_message(msg.chat.id, "Error parsing the json file")
                .await?;
            dialogue.exit().await?;
            return Ok(());
        }
    };
    let previous = bot
        .send_message(msg.chat.id, "Send a new master password")
        .await?;
    dialogue
        .update(State::GetMasterPass(package_handler(
            move |bot, msg, db, dialogue, master_pass| {
                get_master_pass(bot, msg, db, dialogue, master_pass, previous, accounts)
            },
        )))
        .await?;
    Ok(())
}

pub async fn import(bot: Throttle<Bot>, msg: Message, dialogue: MainDialogue) -> crate::Result<()> {
    let previous = bot
        .send_message(
            msg.chat.id,
            "Send a json document with the same format as created by /export",
        )
        .await?;
    dialogue
        .update(State::GetDocument(package_handler(
            move |bot, msg, db, dialogue, ()| get_document(bot, msg, db, dialogue, previous),
        )))
        .await?;
    Ok(())
}
